@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2010 by Digital Mars
+// Copyright (c) 1999-2011 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -104,15 +104,15 @@ int Dsymbol::oneMember(Dsymbol **ps)
  * Same as Dsymbol::oneMember(), but look at an array of Dsymbols.
  */
 
-int Dsymbol::oneMembers(Array *members, Dsymbol **ps)
+int Dsymbol::oneMembers(Dsymbols *members, Dsymbol **ps)
 {
     //printf("Dsymbol::oneMembers() %d\n", members ? members->dim : 0);
     Dsymbol *s = NULL;
 
     if (members)
     {
-        for (int i = 0; i < members->dim; i++)
-        {   Dsymbol *sx = (Dsymbol *)members->data[i];
+        for (size_t i = 0; i < members->dim; i++)
+        {   Dsymbol *sx = (*members)[i];
 
             int x = sx->oneMember(ps);
             //printf("\t[%d] kind %s = %d, s = %p\n", i, sx->kind(), x, *ps);
@@ -197,12 +197,14 @@ const char *Dsymbol::toPrettyChars()
 char *Dsymbol::locToChars()
 {
     OutBuffer buf;
-    char *p;
 
-    Module *m = getModule();
+    if (!loc.filename)  // avoid bug 5861.
+    {
+        Module *m = getModule();
 
-    if (m && m->srcfile)
-        loc.filename = m->srcfile->toChars();
+        if (m && m->srcfile)
+            loc.filename = m->srcfile->toChars();
+    }
     return loc.toChars();
 }
 
@@ -236,17 +238,6 @@ Dsymbol *Dsymbol::pastMixin()
     return s;
 }
 
-TemplateInstance *Dsymbol::inTemplateInstance()
-{
-    for (Dsymbol *parent = this->parent; parent; parent = parent->parent)
-    {
-        TemplateInstance *ti = parent->isTemplateInstance();
-        if (ti)
-            return ti;
-    }
-    return NULL;
-}
-
 /**********************************
  * Use this instead of toParent() when looking for the
  * 'this' pointer of the enclosing function/class.
@@ -260,6 +251,16 @@ Dsymbol *Dsymbol::toParent2()
     return s;
 }
 
+TemplateInstance *Dsymbol::inTemplateInstance()
+{
+    for (Dsymbol *parent = this->parent; parent; parent = parent->parent)
+    {
+        TemplateInstance *ti = parent->isTemplateInstance();
+        if (ti)
+            return ti;
+    }
+    return NULL;
+}
 
 int Dsymbol::isAnonymous()
 {
@@ -286,8 +287,8 @@ void Dsymbol::importAll(Scope *sc)
 /*************************************
  * Does semantic analysis on the public face of declarations.
  */
- 
- void Dsymbol::semantic0(Scope *sc)
+
+void Dsymbol::semantic0(Scope *sc)
 {
 }
 
@@ -422,7 +423,7 @@ void Dsymbol::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
     buf->writestring(toChars());
 }
 
-target_size_t Dsymbol::size(Loc loc)
+unsigned Dsymbol::size(Loc loc)
 {
     error("Dsymbol '%s' has no size\n", toChars());
     return 0;
@@ -481,8 +482,8 @@ LabelDsymbol *Dsymbol::isLabel()                // is this a LabelDsymbol()?
 AggregateDeclaration *Dsymbol::isMember()       // is this a member of an AggregateDeclaration?
 {
     //printf("Dsymbol::isMember() %s\n", toChars());
-      Dsymbol *parent = toParent();
-     //printf("parent is %s %s\n", parent->kind(), parent->toChars());
+    Dsymbol *parent = toParent();
+    //printf("parent is %s %s\n", parent->kind(), parent->toChars());
     return parent ? parent->isAggregateDeclaration() : NULL;
 }
 
@@ -593,12 +594,13 @@ void Dsymbol::checkDeprecated(Loc loc, Scope *sc)
                 return;
         }
 
-        for (; sc; sc = sc->enclosing)
+        for (Scope *sc2 = sc; sc2; sc2 = sc2->enclosing)
         {
-            if (sc->scopesym && sc->scopesym->isDeprecated())
+            if (sc2->scopesym && sc2->scopesym->isDeprecated())
                 return;
-                // If inside a StorageClassDeclaration that is deprecated
-            if (sc->stc & STCdeprecated)
+
+            // If inside a StorageClassDeclaration that is deprecated
+            if (sc2->stc & STCdeprecated)
                 return;
         }
 
@@ -616,6 +618,10 @@ Module *Dsymbol::getModule()
     Dsymbol *s;
 
     //printf("Dsymbol::getModule()\n");
+    TemplateDeclaration *td = getFuncTemplateDecl(this);
+    if (td)
+        return td->getModule();
+
     s = this;
     while (s)
     {
@@ -641,19 +647,19 @@ enum PROT Dsymbol::prot()
  */
 
 
-Array *Dsymbol::arraySyntaxCopy(Array *a)
+Dsymbols *Dsymbol::arraySyntaxCopy(Dsymbols *a)
 {
 
-    Array *b = NULL;
+    Dsymbols *b = NULL;
     if (a)
     {
         b = a->copy();
-        for (int i = 0; i < b->dim; i++)
+        for (size_t i = 0; i < b->dim; i++)
         {
-            Dsymbol *s = (Dsymbol *)b->data[i];
+            Dsymbol *s = (*b)[i];
 
             s = s->syntaxCopy(NULL);
-            b->data[i] = (void *)s;
+            (*b)[i] = s;
         }
     }
     return b;
@@ -667,8 +673,8 @@ Array *Dsymbol::arraySyntaxCopy(Array *a)
 
 void Dsymbol::addComment(unsigned char *comment)
 {
-//    if (comment)
-//      printf("adding comment '%s' to symbol %p '%s'\n", comment, this, toChars());
+    //if (comment)
+        //printf("adding comment '%s' to symbol %p '%s'\n", comment, this, toChars());
 
     if (!this->comment)
         this->comment = comment;
@@ -735,9 +741,9 @@ Dsymbol *ScopeDsymbol::syntaxCopy(Dsymbol *s)
 
 Dsymbol *ScopeDsymbol::search(Loc loc, Identifier *ident, int flags)
 {
-
     //printf("%s->ScopeDsymbol::search(ident='%s', flags=x%x)\n", toChars(), ident->toChars(), flags);
     //if (strcmp(ident->toChars(),"c") == 0) *(char*)0=0;
+
     // Look in symbols declared in this module
     Dsymbol *s = symtab ? symtab->lookup(ident) : NULL;
     if (s)
@@ -747,8 +753,8 @@ Dsymbol *ScopeDsymbol::search(Loc loc, Identifier *ident, int flags)
     else if (imports)
     {
         // Look in imported modules
-        for (int i = 0; i < imports->dim; i++)
-        {   ScopeDsymbol *ss = (ScopeDsymbol *)imports->data[i];
+        for (size_t i = 0; i < imports->dim; i++)
+        {   ScopeDsymbol *ss = (*imports)[i];
             Dsymbol *s2;
 
             // If private import, don't search it
@@ -811,13 +817,11 @@ void ScopeDsymbol::importScope(ScopeDsymbol *s, enum PROT protection)
     if (s != this)
     {
         if (!imports)
-            imports = new Array();
+            imports = new ScopeDsymbols();
         else
         {
-            for (int i = 0; i < imports->dim; i++)
-            {   ScopeDsymbol *ss;
-
-                ss = (ScopeDsymbol *) imports->data[i];
+            for (size_t i = 0; i < imports->dim; i++)
+            {   ScopeDsymbol *ss = (*imports)[i];
                 if (ss == s)                    // if already imported
                 {
                     if (protection > prots[i])
@@ -906,13 +910,13 @@ Dsymbol *ScopeDsymbol::symtabInsert(Dsymbol *s)
  */
 
 #if DMDV2
-size_t ScopeDsymbol::dim(Array *members)
+size_t ScopeDsymbol::dim(Dsymbols *members)
 {
     size_t n = 0;
     if (members)
     {
         for (size_t i = 0; i < members->dim; i++)
-        {   Dsymbol *s = (Dsymbol *)members->data[i];
+        {   Dsymbol *s = (*members)[i];
             AttribDeclaration *a = s->isAttribDeclaration();
 
             if (a)
@@ -936,15 +940,17 @@ size_t ScopeDsymbol::dim(Array *members)
  */
 
 #if DMDV2
-Dsymbol *ScopeDsymbol::getNth(Array *members, size_t nth, size_t *pn)
+Dsymbol *ScopeDsymbol::getNth(Dsymbols *members, size_t nth, size_t *pn)
 {
     if (!members)
         return NULL;
 
     size_t n = 0;
     for (size_t i = 0; i < members->dim; i++)
-    {   Dsymbol *s = (Dsymbol *)members->data[i];
+    {   Dsymbol *s = (*members)[i];
         AttribDeclaration *a = s->isAttribDeclaration();
+        TemplateMixin *tm = s->isTemplateMixin();
+        TemplateInstance *ti = s->isTemplateInstance();
 
         if (a)
         {
@@ -952,6 +958,14 @@ Dsymbol *ScopeDsymbol::getNth(Array *members, size_t nth, size_t *pn)
             if (s)
                 return s;
         }
+        else if (tm)
+        {
+            s = getNth(tm->members, nth - n, &n);
+            if (s)
+                return s;
+        }
+        else if (ti)
+            ;
         else if (n == nth)
             return s;
         else
@@ -1109,35 +1123,25 @@ Dsymbol *ArrayScopeSymbol::search(Loc loc, Identifier *ident, int flags)
          * multiple times, it gets set only once.
          */
         if (!*pvar)             // if not already initialized
-        {   /* Create variable v and set it to the value of $,
-             * which will be a constant.
+        {   /* Create variable v and set it to the value of $
              */
             VarDeclaration *v = new VarDeclaration(loc, Type::tsize_t, Id::dollar, NULL);
-
-            if (ce->op == TOKstring)
-            {   /* It is for a string literal, so the
-                 * length will be a const.
-                 */
-                Expression *e = new IntegerExp(0, ((StringExp *)ce)->len, Type::tsize_t);
-                v->init = new ExpInitializer(0, e);
-                v->storage_class |= STCconst;
-            }
-            else if (ce->op == TOKarrayliteral)
-            {   /* It is for an array literal, so the
-                 * length will be a const.
-                 */
-                Expression *e = new IntegerExp(0, ((ArrayLiteralExp *)ce)->elements->dim, Type::tsize_t);
-                v->init = new ExpInitializer(0, e);
-                v->storage_class |= STCconst;
-            }
-            else if (ce->op == TOKtuple)
+            if (ce->op == TOKtuple)
             {   /* It is for an expression tuple, so the
-                 * length will be a const.
+                 * length will be a compile-time constant.
                  */
                 Expression *e = new IntegerExp(0, ((TupleExp *)ce)->exps->dim, Type::tsize_t);
                 v->init = new ExpInitializer(0, e);
                 v->storage_class |= STCconst;
             }
+            else
+            {   /* For arrays, $ will either be a compile-time constant
+                 * (in which case its value in set during constant-folding),
+                 * or a variable (in which case an expression is created in
+                 * toir.c).
+                 */
+                v->init = new VoidInitializer(0);
+            }
             *pvar = v;
         }
         return (*pvar);
@@ -1151,7 +1155,7 @@ Dsymbol *ArrayScopeSymbol::search(Loc loc, Identifier *ident, int flags)
 DsymbolTable::DsymbolTable()
 {
 #if STRINGTABLE
-     tab = new StringTable;
+    tab = new StringTable;
 #else
     tab = NULL;
 #endif
