@@ -1,5 +1,5 @@
 
-// Copyright (c) 1999-2010 by Digital Mars
+// Copyright (c) 1999-2011 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -11,9 +11,10 @@
    work with the GDC compiler.
 
    Modified by David Friedman, December 2006
+   Updated by Iain Buclaw, January 2011
 */
 
-#define POSIX (linux || __APPLE__ || __FreeBSD__ || __sun&&__SVR4)
+#define POSIX (linux || __APPLE__ || __FreeBSD__ || __OpenBSD__ || __sun&&__SVR4)
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -23,8 +24,18 @@
 #include <stdint.h>
 #include <assert.h>
 
+#if IN_GCC
 #include "gdc_alloca.h"
-#include <string> //for msc_ver
+#else
+#if (defined (__SVR4) && defined (__sun))
+#include <alloca.h>
+#endif
+#endif
+
+#if _MSC_VER ||__MINGW32__
+#include <malloc.h>
+#include <string>
+#endif
 
 #if _WIN32
 #include <windows.h>
@@ -32,7 +43,7 @@
 #include <errno.h>
 #endif
 
-#ifndef _WIN32 //#if linux || __APPLE__ || __FreeBSD__
+#ifndef _WIN32
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -95,6 +106,30 @@ int wcharIsAscii(wchar_t *us, unsigned len)
 #endif
 
 
+#if IN_GCC
+/***********************************
+ * Binary search for p in tab.
+ * Implementation copied from gdc/d-asm-i386.h
+ */
+int binary(const char *p , const char **tab, int high)
+{
+    int low = 0;
+    do
+    {
+        int pos = (low + high) / 2;
+        int cmp = strcmp(p, tab[pos]);
+        if (! cmp)
+            return pos;
+        else if (cmp < 0)
+            high = pos;
+        else
+            low = pos + 1;
+    } while (low != high);
+
+    return -1;
+}
+#endif
+
 /***********************************
  * Compare length-prefixed strings (bstr).
  */
@@ -358,21 +393,21 @@ FileName::FileName(char *path, char *name)
 }
 
 // Split a path into an Array of paths
-Array *FileName::splitPath(const char *path)
+Strings *FileName::splitPath(const char *path)
 {
     char c = 0;                         // unnecessary initializer is for VC /W4
     const char *p;
     OutBuffer buf;
-    Array *array;
+    Strings *array;
 
-    array = new Array();
+    array = new Strings();
     if (path)
     {
         p = path;
         do
         {   char instring = 0;
 
-            while (isspace(*p))         // skip leading whitespace
+            while (isspace((unsigned char)*p))         // skip leading whitespace
                 p++;
             buf.reserve(strlen(p) + 1); // guess size of path
             for (; ; p++)
@@ -479,7 +514,7 @@ hash_t FileName::hashCode()
 
 int FileName::compare(Object *obj)
 {
-        return compare(str, ((FileName *)obj)->str);
+    return compare(str, ((FileName *)obj)->str);
 }
 
 int FileName::compare(const char *name1, const char *name2)
@@ -493,7 +528,7 @@ int FileName::compare(const char *name1, const char *name2)
 
 int FileName::equals(Object *obj)
 {
-        return compare(obj) == 0;
+    return compare(obj) == 0;
 }
 
 int FileName::equals(const char *name1, const char *name2)
@@ -507,14 +542,13 @@ int FileName::equals(const char *name1, const char *name2)
 
 int FileName::absolute(const char *name)
 {
-#ifndef _WIN32
-        return (*name == '/');
-#elif _WIN32
+#if _WIN32
     return (*name == '\\') ||
            (*name == '/')  ||
            (*name && name[1] == ':');
-#else
-    assert(0);
+#endif
+#ifndef _WIN32
+    return (*name == '/');
 #endif
 }
 
@@ -606,7 +640,7 @@ char *FileName::name(const char *str)
                  * Consider ADS separators as part of the file name.
                  */
                 if (e == str + 1 || e == str + len - 1)
-                return e + 1;
+                    return e + 1;
 #endif
             default:
                 if (e == str)
@@ -787,7 +821,7 @@ void FileName::CopyTo(FileName *to)
  *      cwd     if !=0, search current directory before searching path
  */
 
-char *FileName::searchPath(Array *path, const char *name, int cwd)
+char *FileName::searchPath(Strings *path, const char *name, int cwd)
 {
     if (absolute(name))
     {
@@ -803,7 +837,7 @@ char *FileName::searchPath(Array *path, const char *name, int cwd)
 
         for (i = 0; i < path->dim; i++)
         {
-            char *p = (char *)path->data[i];
+            char *p = path->tdata()[i];
             char *n = combine(p, name);
 
             if (exists(n))
@@ -813,6 +847,7 @@ char *FileName::searchPath(Array *path, const char *name, int cwd)
     return NULL;
 }
 
+
 /*************************************
  * Search Path for file in a safe manner.
  *
@@ -826,7 +861,7 @@ char *FileName::searchPath(Array *path, const char *name, int cwd)
  *      !=NULL  mem.malloc'd file name
  */
 
-char *FileName::safeSearchPath(Array *path, const char *name)
+char *FileName::safeSearchPath(Strings *path, const char *name)
 {
 #if _WIN32
     /* Disallow % / \ : and .. in name characters
@@ -863,7 +898,7 @@ char *FileName::safeSearchPath(Array *path, const char *name)
         for (i = 0; i < path->dim; i++)
         {
             char *cname = NULL;
-            char *cpath = canonicalName((char *)path->data[i]);
+            char *cpath = canonicalName(path->tdata()[i]);
             //printf("FileName::safeSearchPath(): name=%s; path=%s; cpath=%s\n",
             //      name, (char *)path->data[i], cpath);
             if (cpath == NULL)
@@ -896,6 +931,7 @@ cont:
 #endif
 }
 
+
 int FileName::exists(const char *name)
 {
 #ifndef _WIN32
@@ -946,14 +982,15 @@ void FileName::ensurePathExists(const char *path)
 #if _WIN32
             if (path[strlen(path) - 1] != '\\')
 #endif
-#if linux
+#ifndef _WIN32
             if (path[strlen(path) - 1] != '\\')
 #endif
             {
                 //printf("mkdir(%s)\n", path);
 #if _WIN32
-                if (mkdir(path))
-#else
+                if (_mkdir(path))
+#endif
+#ifndef _WIN32
                 if (mkdir(path, 0777))
 #endif
                 {
@@ -968,6 +1005,7 @@ void FileName::ensurePathExists(const char *path)
     }
 }
 
+
 /******************************************
  * Return canonical version of name in a malloc'd buffer.
  * This code is high risk.
@@ -1012,6 +1050,7 @@ char *FileName::canonicalName(const char *name)
 #endif
 }
 
+
 /****************************** File ********************************/
 
 File::File(FileName *n)
@@ -1071,7 +1110,7 @@ int File::read()
     //printf("File::read('%s')\n",name);
     fd = open(name, O_RDONLY);
     if (fd == -1)
-    {   result = errno;
+    {
         //printf("\topen error, errno = %d\n",errno);
         goto err1;
     }
@@ -1437,23 +1476,23 @@ void File::remove()
 #endif
 }
 
-Array *File::match(char *n)
+Files *File::match(char *n)
 {
     return match(new FileName(n, 0));
 }
 
-Array *File::match(FileName *n)
+Files *File::match(FileName *n)
 {
 #ifndef _WIN32
     return NULL;
 #elif _WIN32
     HANDLE h;
     WIN32_FIND_DATAA fileinfo;
-    Array *a;
+    Files *a;
     char *c;
     char *name;
 
-    a = new Array();
+    a = new Files();
     c = n->toChars();
     name = n->name();
     h = FindFirstFileA(c,&fileinfo);
@@ -1546,11 +1585,11 @@ OutBuffer::~OutBuffer()
     mem.free(data);
 }
 
-void *OutBuffer::extractData()
+char *OutBuffer::extractData()
 {
-    void *p;
+    char *p;
 
-    p = (void *)data;
+    p = (char *)data;
     data = NULL;
     offset = 0;
     size = 0;
@@ -1752,7 +1791,7 @@ void OutBuffer::writeUTF16(unsigned w)
 void OutBuffer::write4(unsigned w)
 {
     reserve(4);
-    *(uint32_t *)(this->data + offset) = w;
+    *(unsigned *)(this->data + offset) = w;
     offset += 4;
 }
 
@@ -1787,11 +1826,12 @@ void OutBuffer::align(unsigned size)
     fill0(nbytes);
 }
 
+
 ////////////////////////////////////////////////////////////////
 // The compiler shipped with Visual Studio 2005 (and possible
 // other versions) does not support C99 printf format specfiers
 // such as %z and %j
-#if _WIN32 //_MSC_VER
+#if _MSC_VER
 using std::string;
 using std::wstring;
 
@@ -1822,24 +1862,26 @@ void OutBuffer::vprintf(const char *format, va_list args)
     char *p;
     unsigned psize;
     int count;
+
     WORKAROUND_C99_SPECIFIERS_BUG(string, fmt, format);
-    va_list args_copy;
 
     p = buffer;
     psize = sizeof(buffer);
     for (;;)
     {
-        va_copy(args_copy, args);
-#if _WIN32
-        count = _vsnprintf(p,psize,format,args_copy);
+/* MinGW32 supports ANSI with -D__USE_MINGW_ANSI_STDIO */
+#if _WIN32 && !defined( __MINGW32__ )
+        va_list va;
+        va_copy(va, args);
+        count = _vsnprintf(p,psize,format,va);
+        va_end(va);
         if (count != -1)
             break;
         psize *= 2;
-#endif
-#ifndef _WIN32
+#else
         va_list va;
-         va_copy(va, args);
- /*
+        va_copy(va, args);
+/*
   The functions vprintf(), vfprintf(), vsprintf(), vsnprintf()
   are equivalent to the functions printf(), fprintf(), sprintf(),
   snprintf(), respectively, except that they are called with a
@@ -1849,7 +1891,7 @@ void OutBuffer::vprintf(const char *format, va_list args)
   va_end(ap) itself afterwards.
  */
         count = vsnprintf(p,psize,format,va);
-         va_end(va);
+        va_end(va);
         if (count == -1)
             psize *= 2;
         else if (count >= psize)
@@ -1869,23 +1911,28 @@ void OutBuffer::vprintf(const wchar_t *format, va_list args)
     dchar *p;
     unsigned psize;
     int count;
-    va_list args_copy;
+
+    WORKAROUND_C99_SPECIFIERS_BUG(wstring, fmt, format);
 
     p = buffer;
     psize = sizeof(buffer) / sizeof(buffer[0]);
     for (;;)
     {
 #if _WIN32
-        count = _vsnwprintf(p,psize,format,args_copy);
+        va_list va;
+        va_copy(va, args);
+        count = _vsnwprintf(p,psize,format,va);
+        va_end(va);
         if (count != -1)
             break;
         psize *= 2;
 #endif
 #ifndef _WIN32
         va_list va;
-         va_copy(va, args);
+        va_copy(va, args);
         count = vsnwprintf(p,psize,format,va);
-         va_end(va);
+        va_end(va);
+
         if (count == -1)
             psize *= 2;
         else if (count >= psize)
